# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""Tests for application claims."""

import pytest

from azure.confidentialledger.receipt._claims_digest_computation import (
    compute_claims_digest,
)


from ._shared.claims_constants import (
    get_test_application_claims_with_claim_digest_dict,
    get_test_application_claims_with_ledger_entry_dict,
)


@pytest.mark.parametrize(
    "input_claims",
    [
        get_test_application_claims_with_claim_digest_dict(),
        get_test_application_claims_with_ledger_entry_dict(),
        [
            get_test_application_claims_with_claim_digest_dict()[0],
            get_test_application_claims_with_ledger_entry_dict()[0],
        ],
    ],
)
def test_claims_digest_computation_with_valid_claims_returns_successfully(input_claims):
    # Check that compute_claims_digest does not throw any exception
    # with valid claims
    try:
        compute_claims_digest(input_claims)
    except Exception as e:
        pytest.fail(f"compute_claims_digest threw an exception with valid claims {e}")


@pytest.mark.parametrize(
    "input_claims",
    [
        None,
        {},
        [],
        [
            {
                "unknown_key": "unknown_value",
            }
        ],
        [
            {
                "digest": {},
            }
        ],
    ],
)
def test_claims_digest_computation_with_invalid_claim_throws_exception(input_claims):
    # Check that compute_claims_digest throws AssertionError
    with pytest.raises(
        AssertionError,
    ):
        compute_claims_digest(input_claims)


@pytest.mark.parametrize(
    "input_claim_kind,input_claims",
    [
        [None, get_test_application_claims_with_ledger_entry_dict()],
        [None, get_test_application_claims_with_claim_digest_dict()],
        ["invalid_claim_kind", get_test_application_claims_with_ledger_entry_dict()],
        ["invalid_claim_kind", get_test_application_claims_with_claim_digest_dict()],
    ],
)
def test_claims_digest_computation_with_invalid_kind_throws_exception(
    input_claim_kind, input_claims
):
    # Create a claim with an invalid kind
    input_claims[0]["kind"] = input_claim_kind

    # Check that compute_claims_digest throws AssertionError
    with pytest.raises(
        AssertionError,
    ):
        compute_claims_digest(input_claims)


@pytest.mark.parametrize(
    "input_ledger_entry_claim,input_claims",
    [
        [None, get_test_application_claims_with_ledger_entry_dict()],
        [
            {
                "collectionId": "invalid_collection",
                "contents": "invalid_contents",
                "secretKey": "invalid_secret_key",
            },
            get_test_application_claims_with_ledger_entry_dict(),
        ],
        [
            {
                "protocol": get_test_application_claims_with_ledger_entry_dict()[0][
                    "ledgerEntry"
                ]["protocol"],
                "contents": "invalid_contents",
            },
            get_test_application_claims_with_ledger_entry_dict(),
        ],
    ],
)
def test_claims_digest_computation_with_invalid_ledger_entry_claim_throws_exception(
    input_ledger_entry_claim, input_claims
):
    # Create a claim with invalid ledger entry claim
    input_claims[0]["ledgerEntry"] = input_ledger_entry_claim

    # Check that compute_claims_digest throws AssertionError
    with pytest.raises(
        AssertionError,
    ):
        compute_claims_digest(input_claims)


@pytest.mark.parametrize(
    "input_ledger_entry_protocol,input_claims",
    [
        ["invalid_protocol", get_test_application_claims_with_ledger_entry_dict()],
    ],
)
def test_claims_digest_computation_with_invalid_ledger_entry_claim_protocol_throws_exception(
    input_ledger_entry_protocol, input_claims
):
    # Create a claim with invalid ledger entry claim protocol
    input_claims[0]["ledgerEntry"]["protocol"] = input_ledger_entry_protocol

    # Check that compute_claims_digest throws ValueError
    with pytest.raises(
        ValueError,
    ):
        compute_claims_digest(input_claims)


@pytest.mark.parametrize(
    "input_ledger_entry_secret_key,input_claims",
    [
        ["invalid_key", get_test_application_claims_with_ledger_entry_dict()],
    ],
)
def test_claims_digest_computation_with_invalid_ledger_entry_claim_secret_key_throws_exception(
    input_ledger_entry_secret_key, input_claims
):
    # Create a claim with invalid ledger entry claim secret key
    input_claims[0]["ledgerEntry"]["secretKey"] = input_ledger_entry_secret_key

    # Check that compute_claims_digest throws ValueError
    with pytest.raises(
        ValueError,
    ):
        compute_claims_digest(input_claims)


@pytest.mark.parametrize(
    "input_digest_claim,input_claims",
    [
        [None, get_test_application_claims_with_claim_digest_dict()],
        [
            {
                "value": "invalid_value",
                "contents": "invalid_contents",
            },
            get_test_application_claims_with_claim_digest_dict(),
        ],
        [
            {
                "protocol": get_test_application_claims_with_claim_digest_dict()[0][
                    "digest"
                ]["protocol"],
            },
            get_test_application_claims_with_claim_digest_dict(),
        ],
    ],
)
def test_claims_digest_computation_with_invalid_digest_claim_throws_exception(
    input_digest_claim, input_claims
):
    # Create a claim with invalid digest claim
    input_claims[0]["digest"] = input_digest_claim

    # Check that compute_claims_digest throws AssertionError
    with pytest.raises(
        AssertionError,
    ):
        compute_claims_digest(input_claims)


@pytest.mark.parametrize(
    "input_digest_protocol,input_claims",
    [
        ["invalid_protocol", get_test_application_claims_with_claim_digest_dict()],
    ],
)
def test_claims_digest_computation_with_unknown_digest_claim_protocol_does_not_throw_exception(
    input_digest_protocol, input_claims
):
    # Create a claim with invalid digest claim protocol
    input_claims[0]["digest"]["protocol"] = input_digest_protocol

    # Check that compute_claims_digest does not throw any exception
    try:
        compute_claims_digest(input_claims)
    except Exception as e:
        pytest.fail(f"compute_claims_digest threw an exception with valid claims {e}")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: use_collections.py
DESCRIPTION:
    This sample demonstrates how to use collections in Confidential Ledger. In this sample, we write
    ledger entries to different collections. Collections may be used to group semantically or
    logically related ledger entries.
USAGE:
    python use_collections.py
    Set the environment variables with your own values before running the sample:
    1) CONFIDENTIALLEDGER_ENDPOINT - the endpoint of the Confidential Ledger.
"""

import logging
import os
import sys
import tempfile
from typing import Any, Dict

from azure.confidentialledger import ConfidentialLedgerClient
from azure.confidentialledger.certificate import (
    ConfidentialLedgerCertificateClient,
)
from azure.identity import DefaultAzureCredential


logging.basicConfig(level=logging.ERROR)
LOG = logging.getLogger()


def main():
    # Set the values of the client ID, tenant ID, and client secret of the AAD application as
    # environment variables:
    #   AZURE_CLIENT_ID, AZURE_TENANT_ID, AZURE_CLIENT_SECRET, CONFIDENTIALLEDGER_ENDPOINT
    try:
        ledger_endpoint = os.environ["CONFIDENTIALLEDGER_ENDPOINT"]
    except KeyError:
        LOG.error(
            "Missing environment variable 'CONFIDENTIALLEDGER_ENDPOINT' - "
            "please set it before running the example"
        )
        sys.exit(1)

    # Under the current URI format, the ledger id is the first part of the ledger endpoint.
    # i.e. https://<ledger id>.confidential-ledger.azure.com
    ledger_id = ledger_endpoint.replace("https://", "").split(".")[0]

    identity_service_client = ConfidentialLedgerCertificateClient()  # type: ignore[call-arg]
    ledger_certificate = identity_service_client.get_ledger_identity(ledger_id)

    # The Confidential Ledger's TLS certificate must be written to a file to be used by the
    # ConfidentialLedgerClient. Here, we write it to a temporary file so that is is cleaned up
    # automatically when the program exits.
    with tempfile.TemporaryDirectory() as tempdir:
        ledger_cert_file = os.path.join(tempdir, f"{ledger_id}.pem")
        with open(ledger_cert_file, "w") as outfile:
            outfile.write(ledger_certificate["ledgerTlsCertificate"])

        print(
            f"Ledger certificate has been written to {ledger_cert_file}. "
            "It will be deleted when the script completes."
        )

        # Build a client through AAD
        ledger_client = ConfidentialLedgerClient(
            ledger_endpoint,
            credential=DefaultAzureCredential(),
            ledger_certificate_path=ledger_cert_file,
        )

        print("This Confidential Ledger will contain messages from different senders.")
        print(
            "We will group ledger entries by the sender of the message. For all client methods "
            "that take an optional 'collection_id' parameter, if none is provided, a "
            "service-assigned, default collection id will be assigned."
        )

        tids: Dict[Any, Dict[str, Any]] = {}
        senders = [None, "Alice", "Bob"]
        for msg_idx in range(3):
            for sender in senders:
                if sender is None:
                    msg = f"My message {msg_idx}"
                else:
                    msg = f"{sender}'s message {msg_idx}"

                post_poller = ledger_client.begin_create_ledger_entry(  # type: ignore[attr-defined]
                    entry={"contents": msg}, collection_id=sender,
                )
                post_result = post_poller.result()

                if sender is None:
                    print(
                        f"Wrote '{msg}' to the default collection at {post_result['transactionId']}"
                    )
                else:
                    print(f"Wrote '{msg}' to collection {sender} at {post_result['transactionId']}")

                if sender not in tids:
                    tids[sender] = {}
                    tids[sender]["first"] = post_result["transactionId"]

                tids[sender]["last"] = post_result["transactionId"]

        print("Let's retrieve the latest entry in each collection")
        for sender in senders:
            current_entry = ledger_client.get_current_ledger_entry()
            
            output = "Current entry in {0} is {1}"
            print(
                output.format(
                    "default collection" if sender is None else f"{sender}'s collection",
                    current_entry["contents"],
                )
            )

        print("Let's retrieve the first entry in each collection")
        for sender in senders:
            get_poller = ledger_client.begin_get_ledger_entry(  # type: ignore[attr-defined]
                tids[sender]["first"],
                collection_id=sender
            )
            first_entry = get_poller.result()
            
            output = "First entry in {0} is {1}"
            print(
                output.format(
                    "default collection" if sender is None else f"{sender}'s collection",
                    first_entry["entry"]["contents"],
                )
            )

        print("Let's retrieve get all the entries in each collection")
        for sender in senders:
            entries_list = ledger_client.list_ledger_entries(
                collection_id=sender,
                from_transaction_id=tids[sender]["first"],
                to_transaction_id=tids[sender]["last"],
            )
            
            for entry in entries_list:
                output = "Entry in {0}: {1}"
                print(
                    output.format(
                        "default collection" if sender is None else f"{sender}'s collection",
                        entry,
                    )
                )

        collection_ids = []
        collections = ledger_client.list_collections()
        for collection in collections:
            collection_ids.append(collection["collectionId"])

        print(
            "In conclusion, these are all the collections in the Confidential Ledger:\n" +
            "\n\t".join(collection_ids)
        )


if __name__ == "__main__":
    main()



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

from azure.confidentialledger.receipt._claims_models import (
    ApplicationClaim,
    LedgerEntryClaim,
    ClaimDigest,
)


def get_test_application_claims_with_ledger_entry_dict():
    return [
        {
            "kind": "LedgerEntry",
            "ledgerEntry": {
                "collectionId": "subledger:0",
                "contents": "Hello world",
                "protocol": "LedgerEntryV1",
                "secretKey": "Jde/VvaIfyrjQ/B19P+UJCBwmcrgN7sERStoyHnYO0M=",
            },
        }
    ]


def get_test_application_claims_with_ledger_entry():
    return [
        ApplicationClaim(
            kind="LedgerEntry",
            ledgerEntry=LedgerEntryClaim(
                collectionId="subledger:0",
                contents="Hello world",
                protocol="LedgerEntryV1",
                secretKey="Jde/VvaIfyrjQ/B19P+UJCBwmcrgN7sERStoyHnYO0M=",
            ),
        )
    ]


def get_test_application_claims_with_claim_digest_dict():
    return [
        {
            "kind": "ClaimDigest",
            "digest": {
                "protocol": "LedgerEntryV1",
                "value": "feb516ef1f903c64f1e388d9ee9fde11f64d1e2bc170248828c9eab9894f9ab9",
            },
        }
    ]


def get_test_application_claims_with_claim_digest():
    return [
        ApplicationClaim(
            kind="ClaimDigest",
            digest=ClaimDigest(
                protocol="LedgerEntryV1",
                value="feb516ef1f903c64f1e388d9ee9fde11f64d1e2bc170248828c9eab9894f9ab9",
            ),
        )
    ]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: get_receipt_async.py
DESCRIPTION:
    This sample demonstrates how to retrieve Confidential Ledger receipts. In this sample, we write
    a ledger entry and retrieve a receipt certifying that it was written correctly. 
USAGE:
    python get_receipt_async.py
    Set the environment variables with your own values before running the sample:
    1) CONFIDENTIALLEDGER_ENDPOINT - the endpoint of the Confidential Ledger.
"""

import asyncio
import logging
import os
import sys
import tempfile

from azure.confidentialledger.aio import ConfidentialLedgerClient
from azure.confidentialledger.certificate.aio import (
    ConfidentialLedgerCertificateClient,
)
from azure.core.exceptions import HttpResponseError
from azure.identity.aio import DefaultAzureCredential


logging.basicConfig(level=logging.ERROR)
LOG = logging.getLogger()


async def main():
    # Set the values of the client ID, tenant ID, and client secret of the AAD application as
    # environment variables:
    #   AZURE_CLIENT_ID, AZURE_TENANT_ID, AZURE_CLIENT_SECRET, CONFIDENTIALLEDGER_ENDPOINT
    try:
        ledger_endpoint = os.environ["CONFIDENTIALLEDGER_ENDPOINT"]
    except KeyError:
        LOG.error(
            "Missing environment variable 'CONFIDENTIALLEDGER_ENDPOINT' - "
            "please set it before running the example"
        )
        sys.exit(1)

    # Under the current URI format, the ledger id is the first part of the ledger endpoint.
    # i.e. https://<ledger id>.confidential-ledger.azure.com
    ledger_id = ledger_endpoint.replace("https://", "").split(".")[0]

    identity_service_client = ConfidentialLedgerCertificateClient()  # type: ignore[call-arg]
    async with identity_service_client:
        ledger_certificate = await identity_service_client.get_ledger_identity(
            ledger_id
        )

    # The Confidential Ledger's TLS certificate must be written to a file to be used by the
    # ConfidentialLedgerClient. Here, we write it to a temporary file so that is is cleaned up
    # automatically when the program exits.
    with tempfile.TemporaryDirectory() as tempdir:
        ledger_cert_file = os.path.join(tempdir, f"{ledger_id}.pem")
        with open(ledger_cert_file, "w") as outfile:
            outfile.write(ledger_certificate["ledgerTlsCertificate"])

        print(
            f"Ledger certificate has been written to {ledger_cert_file}. "
            "It will be deleted when the script completes."
        )

        # Build a client through AAD
        credential = DefaultAzureCredential()
        ledger_client = ConfidentialLedgerClient(
            ledger_endpoint,
            credential=credential,
            ledger_certificate_path=ledger_cert_file,
        )

        # Using the async objects as a context manager ensures they are properly closed after use.
        async with credential:
            async with ledger_client:
                # Write a ledger entry.
                try:
                    entry_contents = "Hello world!"
                    post_poller = await ledger_client.begin_create_ledger_entry(  # type: ignore[attr-defined]
                        {"contents": entry_contents}
                    )
                    post_entry_result = await post_poller.result()
                    transaction_id = post_entry_result["transactionId"]
                    print(f"Wrote '{entry_contents}' to the ledger at transaction {transaction_id}.")
                except HttpResponseError as e:
                    print("Request failed: {}".format(e.response.json()))  # type: ignore[union-attr]
                    raise

                # Get a receipt for a ledger entry.
                # A receipt can be retrieved for any transaction id to provide cryptographic proof
                # of the contents of the transaction.
                try:
                    print(
                        f"Retrieving a receipt for {transaction_id}. The receipt may be used to "
                        "cryptographically verify the contents of the transaction."
                    )
                    print(
                        "For more information about receipts, please see "
                        "https://microsoft.github.io/CCF/main/audit/receipts.html#receipts"
                    )
                    get_receipt_poller = await ledger_client.begin_get_receipt(transaction_id)  # type: ignore[attr-defined]
                    get_receipt_result = await get_receipt_poller.result()
                    print(
                        f'Receipt for transaction id {transaction_id}: {get_receipt_result}'
                    )
                except HttpResponseError as e:
                    print("Request failed: {}".format(e.response.json()))  # type: ignore[union-attr]
                    raise


if __name__ == "__main__":
    asyncio.run(main())



# pylint: disable=line-too-long,useless-suppression,too-many-lines
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
from collections.abc import MutableMapping
from io import IOBase
from typing import Any, AsyncIterable, Callable, Dict, IO, List, Optional, TypeVar, Union, cast, overload
import urllib.parse

from azure.core import AsyncPipelineClient
from azure.core.async_paging import AsyncItemPaged, AsyncList
from azure.core.exceptions import (
    ClientAuthenticationError,
    HttpResponseError,
    ResourceExistsError,
    ResourceNotFoundError,
    ResourceNotModifiedError,
    map_error,
)
from azure.core.pipeline import PipelineResponse
from azure.core.rest import AsyncHttpResponse, HttpRequest
from azure.core.tracing.decorator import distributed_trace
from azure.core.tracing.decorator_async import distributed_trace_async
from azure.core.utils import case_insensitive_dict

from ..._operations._operations import (
    build_confidential_ledger_create_ledger_entry_request,
    build_confidential_ledger_create_or_update_ledger_user_request,
    build_confidential_ledger_create_or_update_user_request,
    build_confidential_ledger_create_user_defined_endpoint_request,
    build_confidential_ledger_create_user_defined_function_request,
    build_confidential_ledger_create_user_defined_role_request,
    build_confidential_ledger_delete_ledger_user_request,
    build_confidential_ledger_delete_user_defined_function_request,
    build_confidential_ledger_delete_user_defined_role_request,
    build_confidential_ledger_delete_user_request,
    build_confidential_ledger_execute_user_defined_function_request,
    build_confidential_ledger_get_constitution_request,
    build_confidential_ledger_get_current_ledger_entry_request,
    build_confidential_ledger_get_enclave_quotes_request,
    build_confidential_ledger_get_ledger_entry_request,
    build_confidential_ledger_get_ledger_user_request,
    build_confidential_ledger_get_receipt_request,
    build_confidential_ledger_get_runtime_options_request,
    build_confidential_ledger_get_transaction_status_request,
    build_confidential_ledger_get_user_defined_endpoint_request,
    build_confidential_ledger_get_user_defined_endpoints_module_request,
    build_confidential_ledger_get_user_defined_function_request,
    build_confidential_ledger_get_user_defined_role_request,
    build_confidential_ledger_get_user_request,
    build_confidential_ledger_list_collections_request,
    build_confidential_ledger_list_consortium_members_request,
    build_confidential_ledger_list_ledger_entries_request,
    build_confidential_ledger_list_ledger_users_request,
    build_confidential_ledger_list_user_defined_functions_request,
    build_confidential_ledger_list_users_request,
    build_confidential_ledger_update_runtime_options_request,
    build_confidential_ledger_update_user_defined_role_request,
)
from ..._utils.utils import ClientMixinABC
from .._configuration import ConfidentialLedgerClientConfiguration

JSON = MutableMapping[str, Any]
T = TypeVar("T")
ClsType = Optional[Callable[[PipelineResponse[HttpRequest, AsyncHttpResponse], T, Dict[str, Any]], Any]]


class ConfidentialLedgerClientOperationsMixin(  # pylint: disable=too-many-public-methods
    ClientMixinABC[AsyncPipelineClient[HttpRequest, AsyncHttpResponse], ConfidentialLedgerClientConfiguration]
):

    @distributed_trace_async
    async def get_constitution(self, **kwargs: Any) -> JSON:
        """Gets the constitution used for governance.

        The constitution is a script that assesses and applies proposals from consortium members.

        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "digest": "str",
                    "script": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_constitution_request(
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace
    def list_consortium_members(self, **kwargs: Any) -> AsyncIterable[JSON]:
        """Lists the consortium members.

        Consortium members can manage the Confidential Ledger.

        :return: An iterator like instance of JSON object
        :rtype: ~azure.core.async_paging.AsyncItemPaged[JSON]
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "certificate": "str",
                    "id": "str"
                }
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                _request = build_confidential_ledger_list_consortium_members_request(
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                _request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            return _request

        async def extract_data(pipeline_response):
            deserialized = pipeline_response.http_response.json()
            list_of_elem = deserialized.get("members", [])
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.get("nextLink") or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            _request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
                _request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                raise HttpResponseError(response=response)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @distributed_trace_async
    async def get_enclave_quotes(self, **kwargs: Any) -> JSON:
        """Gets quotes for all nodes of the Confidential Ledger.

        A quote is an SGX enclave measurement that can be used to verify the validity of a node and its
        enclave.

        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "currentNodeId": "str",
                    "enclaveQuotes": {
                        "str": {
                            "nodeId": "str",
                            "quoteVersion": "str",
                            "raw": "str",
                            "mrenclave": "str"
                        }
                    }
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_enclave_quotes_request(
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace
    def list_collections(self, **kwargs: Any) -> AsyncIterable[JSON]:
        """Retrieves a list of collection ids present in the Confidential Ledger.

        Collection ids are user-created collections of ledger entries.

        :return: An iterator like instance of JSON object
        :rtype: ~azure.core.async_paging.AsyncItemPaged[JSON]
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "collectionId": "str"
                }
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                _request = build_confidential_ledger_list_collections_request(
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                _request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            return _request

        async def extract_data(pipeline_response):
            deserialized = pipeline_response.http_response.json()
            list_of_elem = deserialized.get("collections", [])
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.get("nextLink") or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            _request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
                _request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                raise HttpResponseError(response=response)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @distributed_trace
    def list_ledger_entries(
        self,
        *,
        collection_id: Optional[str] = None,
        from_transaction_id: Optional[str] = None,
        to_transaction_id: Optional[str] = None,
        tag: Optional[str] = None,
        **kwargs: Any
    ) -> AsyncIterable[JSON]:
        """Gets ledger entries from a collection corresponding to a range.

        A collection id may optionally be specified. Only entries in the specified (or default)
        collection will be returned.

        :keyword collection_id: The collection id. Default value is None.
        :paramtype collection_id: str
        :keyword from_transaction_id: Specify the first transaction ID in a range. Default value is
         None.
        :paramtype from_transaction_id: str
        :keyword to_transaction_id: Specify the last transaction ID in a range. Default value is None.
        :paramtype to_transaction_id: str
        :keyword tag: Single tag. Default value is None.
        :paramtype tag: str
        :return: An iterator like instance of JSON object
        :rtype: ~azure.core.async_paging.AsyncItemPaged[JSON]
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "contents": "str",
                    "collectionId": "str",
                    "postHooks": [
                        {
                            "functionId": "str",
                            "properties": {
                                "arguments": [
                                    "str"
                                ],
                                "exportedFunctionName": "str",
                                "runtimeOptions": {
                                    "log_exception_details": bool,
                                    "max_cached_interpreters": 0,
                                    "max_execution_time_ms": 0,
                                    "max_heap_bytes": 0,
                                    "max_stack_bytes": 0,
                                    "return_exception_details": bool
                                }
                            }
                        }
                    ],
                    "preHooks": [
                        {
                            "functionId": "str",
                            "properties": {
                                "arguments": [
                                    "str"
                                ],
                                "exportedFunctionName": "str",
                                "runtimeOptions": {
                                    "log_exception_details": bool,
                                    "max_cached_interpreters": 0,
                                    "max_execution_time_ms": 0,
                                    "max_heap_bytes": 0,
                                    "max_stack_bytes": 0,
                                    "return_exception_details": bool
                                }
                            }
                        }
                    ],
                    "transactionId": "str"
                }
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                _request = build_confidential_ledger_list_ledger_entries_request(
                    collection_id=collection_id,
                    from_transaction_id=from_transaction_id,
                    to_transaction_id=to_transaction_id,
                    tag=tag,
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                _request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            return _request

        async def extract_data(pipeline_response):
            deserialized = pipeline_response.http_response.json()
            list_of_elem = deserialized.get("entries", [])
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.get("nextLink") or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            _request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
                _request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                raise HttpResponseError(response=response)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @overload
    async def create_ledger_entry(
        self,
        entry: JSON,
        *,
        collection_id: Optional[str] = None,
        tags: Optional[str] = None,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> JSON:
        """Writes a ledger entry.

        A collection id may optionally be specified.

        :param entry: Ledger entry. Required.
        :type entry: JSON
        :keyword collection_id: The collection id. Default value is None.
        :paramtype collection_id: str
        :keyword tags: Comma separated tags. Default value is None.
        :paramtype tags: str
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                entry = {
                    "contents": "str",
                    "collectionId": "str",
                    "postHooks": [
                        {
                            "functionId": "str",
                            "properties": {
                                "arguments": [
                                    "str"
                                ],
                                "exportedFunctionName": "str",
                                "runtimeOptions": {
                                    "log_exception_details": bool,
                                    "max_cached_interpreters": 0,
                                    "max_execution_time_ms": 0,
                                    "max_heap_bytes": 0,
                                    "max_stack_bytes": 0,
                                    "return_exception_details": bool
                                }
                            }
                        }
                    ],
                    "preHooks": [
                        {
                            "functionId": "str",
                            "properties": {
                                "arguments": [
                                    "str"
                                ],
                                "exportedFunctionName": "str",
                                "runtimeOptions": {
                                    "log_exception_details": bool,
                                    "max_cached_interpreters": 0,
                                    "max_execution_time_ms": 0,
                                    "max_heap_bytes": 0,
                                    "max_stack_bytes": 0,
                                    "return_exception_details": bool
                                }
                            }
                        }
                    ],
                    "transactionId": "str"
                }

                # response body for status code(s): 200
                response == {
                    "collectionId": "str"
                }
        """

    @overload
    async def create_ledger_entry(
        self,
        entry: IO[bytes],
        *,
        collection_id: Optional[str] = None,
        tags: Optional[str] = None,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> JSON:
        """Writes a ledger entry.

        A collection id may optionally be specified.

        :param entry: Ledger entry. Required.
        :type entry: IO[bytes]
        :keyword collection_id: The collection id. Default value is None.
        :paramtype collection_id: str
        :keyword tags: Comma separated tags. Default value is None.
        :paramtype tags: str
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "collectionId": "str"
                }
        """

    @distributed_trace_async
    async def create_ledger_entry(
        self,
        entry: Union[JSON, IO[bytes]],
        *,
        collection_id: Optional[str] = None,
        tags: Optional[str] = None,
        **kwargs: Any
    ) -> JSON:
        """Writes a ledger entry.

        A collection id may optionally be specified.

        :param entry: Ledger entry. Is either a JSON type or a IO[bytes] type. Required.
        :type entry: JSON or IO[bytes]
        :keyword collection_id: The collection id. Default value is None.
        :paramtype collection_id: str
        :keyword tags: Comma separated tags. Default value is None.
        :paramtype tags: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                entry = {
                    "contents": "str",
                    "collectionId": "str",
                    "postHooks": [
                        {
                            "functionId": "str",
                            "properties": {
                                "arguments": [
                                    "str"
                                ],
                                "exportedFunctionName": "str",
                                "runtimeOptions": {
                                    "log_exception_details": bool,
                                    "max_cached_interpreters": 0,
                                    "max_execution_time_ms": 0,
                                    "max_heap_bytes": 0,
                                    "max_stack_bytes": 0,
                                    "return_exception_details": bool
                                }
                            }
                        }
                    ],
                    "preHooks": [
                        {
                            "functionId": "str",
                            "properties": {
                                "arguments": [
                                    "str"
                                ],
                                "exportedFunctionName": "str",
                                "runtimeOptions": {
                                    "log_exception_details": bool,
                                    "max_cached_interpreters": 0,
                                    "max_execution_time_ms": 0,
                                    "max_heap_bytes": 0,
                                    "max_stack_bytes": 0,
                                    "return_exception_details": bool
                                }
                            }
                        }
                    ],
                    "transactionId": "str"
                }

                # response body for status code(s): 200
                response == {
                    "collectionId": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[JSON] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(entry, (IOBase, bytes)):
            _content = entry
        else:
            _json = entry

        _request = build_confidential_ledger_create_ledger_entry_request(
            collection_id=collection_id,
            tags=tags,
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        response_headers = {}
        response_headers["x-ms-ccf-transaction-id"] = self._deserialize(
            "str", response.headers.get("x-ms-ccf-transaction-id")
        )

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), response_headers)  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace_async
    async def get_ledger_entry(
        self, transaction_id: str, *, collection_id: Optional[str] = None, **kwargs: Any
    ) -> JSON:
        """Gets the ledger entry at the specified transaction id. A collection id may optionally be
        specified to indicate the collection from which to fetch the value.

        To return older ledger entries, the relevant sections of the ledger must be read from disk and
        validated. To prevent blocking within the enclave, the response will indicate whether the entry
        is ready and part of the response, or if the loading is still ongoing.

        :param transaction_id: Identifies a write transaction. Required.
        :type transaction_id: str
        :keyword collection_id: The collection id. Default value is None.
        :paramtype collection_id: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "state": "str",
                    "entry": {
                        "contents": "str",
                        "collectionId": "str",
                        "postHooks": [
                            {
                                "functionId": "str",
                                "properties": {
                                    "arguments": [
                                        "str"
                                    ],
                                    "exportedFunctionName": "str",
                                    "runtimeOptions": {
                                        "log_exception_details": bool,
                                        "max_cached_interpreters": 0,
                                        "max_execution_time_ms": 0,
                                        "max_heap_bytes": 0,
                                        "max_stack_bytes": 0,
                                        "return_exception_details": bool
                                    }
                                }
                            }
                        ],
                        "preHooks": [
                            {
                                "functionId": "str",
                                "properties": {
                                    "arguments": [
                                        "str"
                                    ],
                                    "exportedFunctionName": "str",
                                    "runtimeOptions": {
                                        "log_exception_details": bool,
                                        "max_cached_interpreters": 0,
                                        "max_execution_time_ms": 0,
                                        "max_heap_bytes": 0,
                                        "max_stack_bytes": 0,
                                        "return_exception_details": bool
                                    }
                                }
                            }
                        ],
                        "transactionId": "str"
                    }
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_ledger_entry_request(
            transaction_id=transaction_id,
            collection_id=collection_id,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace_async
    async def get_receipt(self, transaction_id: str, **kwargs: Any) -> JSON:
        """Gets a receipt certifying ledger contents at a particular transaction id.

        Gets a receipt certifying ledger contents at a particular transaction id.

        :param transaction_id: Identifies a write transaction. Required.
        :type transaction_id: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "state": "str",
                    "transactionId": "str",
                    "applicationClaims": [
                        {
                            "kind": "str",
                            "digest": {
                                "protocol": "str",
                                "value": "str"
                            },
                            "ledgerEntry": {
                                "protocol": "str",
                                "collectionId": "str",
                                "contents": "str",
                                "secretKey": "str"
                            }
                        }
                    ],
                    "receipt": {
                        "nodeId": "str",
                        "proof": [
                            {
                                "left": "str",
                                "right": "str"
                            }
                        ],
                        "signature": "str",
                        "cert": "str",
                        "leaf": "str",
                        "leafComponents": {
                            "claimsDigest": "str",
                            "commitEvidence": "str",
                            "writeSetDigest": "str"
                        },
                        "root": "str",
                        "serviceEndorsements": [
                            "str"
                        ]
                    }
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_receipt_request(
            transaction_id=transaction_id,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace_async
    async def get_transaction_status(self, transaction_id: str, **kwargs: Any) -> JSON:
        """Gets the status of an entry identified by a transaction id.

        Gets the status of an entry identified by a transaction id.

        :param transaction_id: Identifies a write transaction. Required.
        :type transaction_id: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "state": "str",
                    "transactionId": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_transaction_status_request(
            transaction_id=transaction_id,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace_async
    async def get_current_ledger_entry(self, *, collection_id: Optional[str] = None, **kwargs: Any) -> JSON:
        """Gets the current value available in the ledger.

        A collection id may optionally be specified.

        :keyword collection_id: The collection id. Default value is None.
        :paramtype collection_id: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "contents": "str",
                    "collectionId": "str",
                    "postHooks": [
                        {
                            "functionId": "str",
                            "properties": {
                                "arguments": [
                                    "str"
                                ],
                                "exportedFunctionName": "str",
                                "runtimeOptions": {
                                    "log_exception_details": bool,
                                    "max_cached_interpreters": 0,
                                    "max_execution_time_ms": 0,
                                    "max_heap_bytes": 0,
                                    "max_stack_bytes": 0,
                                    "return_exception_details": bool
                                }
                            }
                        }
                    ],
                    "preHooks": [
                        {
                            "functionId": "str",
                            "properties": {
                                "arguments": [
                                    "str"
                                ],
                                "exportedFunctionName": "str",
                                "runtimeOptions": {
                                    "log_exception_details": bool,
                                    "max_cached_interpreters": 0,
                                    "max_execution_time_ms": 0,
                                    "max_heap_bytes": 0,
                                    "max_stack_bytes": 0,
                                    "return_exception_details": bool
                                }
                            }
                        }
                    ],
                    "transactionId": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_current_ledger_entry_request(
            collection_id=collection_id,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace
    def list_users(self, **kwargs: Any) -> AsyncIterable[JSON]:
        """Gets details on a list of users.

        All users' object IDs and single role per user will be returned.

        :return: An iterator like instance of JSON object
        :rtype: ~azure.core.async_paging.AsyncItemPaged[JSON]
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "assignedRole": "str",
                    "userId": "str"
                }
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                _request = build_confidential_ledger_list_users_request(
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                _request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            return _request

        async def extract_data(pipeline_response):
            deserialized = pipeline_response.http_response.json()
            list_of_elem = deserialized.get("ledgerUsers", [])
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.get("nextLink") or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            _request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
                _request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                raise HttpResponseError(response=response)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @distributed_trace
    def list_ledger_users(self, **kwargs: Any) -> AsyncIterable[JSON]:
        """Gets details on a list of users.

        All users' object IDs and multiple roles will be returned.

        :return: An iterator like instance of JSON object
        :rtype: ~azure.core.async_paging.AsyncItemPaged[JSON]
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "assignedRoles": [
                        "str"
                    ],
                    "userId": "str"
                }
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                _request = build_confidential_ledger_list_ledger_users_request(
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                _request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            return _request

        async def extract_data(pipeline_response):
            deserialized = pipeline_response.http_response.json()
            list_of_elem = deserialized.get("ledgerUsers", [])
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.get("nextLink") or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            _request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
                _request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                raise HttpResponseError(response=response)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @distributed_trace_async
    async def delete_user(self, user_id: str, **kwargs: Any) -> None:
        """Deletes a user from the Confidential Ledger.

        Deletes a user from the Confidential Ledger.

        :param user_id: The user id, either an AAD object ID or certificate fingerprint. Required.
        :type user_id: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[None] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_delete_user_request(
            user_id=user_id,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [204]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @distributed_trace_async
    async def get_user(self, user_id: str, **kwargs: Any) -> JSON:
        """Gets a user.

        Gets a user.

        :param user_id: The user id, either an AAD object ID or certificate fingerprint. Required.
        :type user_id: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "assignedRole": "str",
                    "userId": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_user_request(
            user_id=user_id,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @overload
    async def create_or_update_user(
        self, user_id: str, user_details: JSON, *, content_type: str = "application/merge-patch+json", **kwargs: Any
    ) -> JSON:
        """Adds a user or updates a user's fields.

        A JSON merge patch is applied for existing users.

        :param user_id: The user id, either an AAD object ID or certificate fingerprint. Required.
        :type user_id: str
        :param user_details: Details about a Confidential Ledger user. Required.
        :type user_details: JSON
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/merge-patch+json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                user_details = {
                    "assignedRole": "str",
                    "userId": "str"
                }

                # response body for status code(s): 200
                response == {
                    "assignedRole": "str",
                    "userId": "str"
                }
        """

    @overload
    async def create_or_update_user(
        self,
        user_id: str,
        user_details: IO[bytes],
        *,
        content_type: str = "application/merge-patch+json",
        **kwargs: Any
    ) -> JSON:
        """Adds a user or updates a user's fields.

        A JSON merge patch is applied for existing users.

        :param user_id: The user id, either an AAD object ID or certificate fingerprint. Required.
        :type user_id: str
        :param user_details: Details about a Confidential Ledger user. Required.
        :type user_details: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/merge-patch+json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "assignedRole": "str",
                    "userId": "str"
                }
        """

    @distributed_trace_async
    async def create_or_update_user(self, user_id: str, user_details: Union[JSON, IO[bytes]], **kwargs: Any) -> JSON:
        """Adds a user or updates a user's fields.

        A JSON merge patch is applied for existing users.

        :param user_id: The user id, either an AAD object ID or certificate fingerprint. Required.
        :type user_id: str
        :param user_details: Details about a Confidential Ledger user. Is either a JSON type or a
         IO[bytes] type. Required.
        :type user_details: JSON or IO[bytes]
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                user_details = {
                    "assignedRole": "str",
                    "userId": "str"
                }

                # response body for status code(s): 200
                response == {
                    "assignedRole": "str",
                    "userId": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[JSON] = kwargs.pop("cls", None)

        content_type = content_type or "application/merge-patch+json"
        _json = None
        _content = None
        if isinstance(user_details, (IOBase, bytes)):
            _content = user_details
        else:
            _json = user_details

        _request = build_confidential_ledger_create_or_update_user_request(
            user_id=user_id,
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace_async
    async def delete_ledger_user(self, user_id: str, **kwargs: Any) -> None:
        """Deletes a user with multiple roles from the Confidential Ledger.

        Deletes a user with multiple roles from the Confidential Ledger.

        :param user_id: The user id, either an AAD object ID or certificate fingerprint. Required.
        :type user_id: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[None] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_delete_ledger_user_request(
            user_id=user_id,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [204]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @distributed_trace_async
    async def get_ledger_user(self, user_id: str, **kwargs: Any) -> JSON:
        """Gets a user with multiple roles.

        Gets a user with multiple roles.

        :param user_id: The user id, either an AAD object ID or certificate fingerprint. Required.
        :type user_id: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "assignedRoles": [
                        "str"
                    ],
                    "userId": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_ledger_user_request(
            user_id=user_id,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @overload
    async def create_or_update_ledger_user(
        self,
        user_id: str,
        user_multiple_roles: JSON,
        *,
        content_type: str = "application/merge-patch+json",
        **kwargs: Any
    ) -> JSON:
        """Adds a user or updates a user's fields.

        A JSON merge patch is applied for existing users.

        :param user_id: The user id, either an AAD object ID or certificate fingerprint. Required.
        :type user_id: str
        :param user_multiple_roles: Details about a Confidential Ledger user with multiple roles.
         Required.
        :type user_multiple_roles: JSON
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/merge-patch+json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                user_multiple_roles = {
                    "assignedRoles": [
                        "str"
                    ],
                    "userId": "str"
                }

                # response body for status code(s): 200
                response == {
                    "assignedRoles": [
                        "str"
                    ],
                    "userId": "str"
                }
        """

    @overload
    async def create_or_update_ledger_user(
        self,
        user_id: str,
        user_multiple_roles: IO[bytes],
        *,
        content_type: str = "application/merge-patch+json",
        **kwargs: Any
    ) -> JSON:
        """Adds a user or updates a user's fields.

        A JSON merge patch is applied for existing users.

        :param user_id: The user id, either an AAD object ID or certificate fingerprint. Required.
        :type user_id: str
        :param user_multiple_roles: Details about a Confidential Ledger user with multiple roles.
         Required.
        :type user_multiple_roles: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/merge-patch+json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "assignedRoles": [
                        "str"
                    ],
                    "userId": "str"
                }
        """

    @distributed_trace_async
    async def create_or_update_ledger_user(
        self, user_id: str, user_multiple_roles: Union[JSON, IO[bytes]], **kwargs: Any
    ) -> JSON:
        """Adds a user or updates a user's fields.

        A JSON merge patch is applied for existing users.

        :param user_id: The user id, either an AAD object ID or certificate fingerprint. Required.
        :type user_id: str
        :param user_multiple_roles: Details about a Confidential Ledger user with multiple roles. Is
         either a JSON type or a IO[bytes] type. Required.
        :type user_multiple_roles: JSON or IO[bytes]
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                user_multiple_roles = {
                    "assignedRoles": [
                        "str"
                    ],
                    "userId": "str"
                }

                # response body for status code(s): 200
                response == {
                    "assignedRoles": [
                        "str"
                    ],
                    "userId": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[JSON] = kwargs.pop("cls", None)

        content_type = content_type or "application/merge-patch+json"
        _json = None
        _content = None
        if isinstance(user_multiple_roles, (IOBase, bytes)):
            _content = user_multiple_roles
        else:
            _json = user_multiple_roles

        _request = build_confidential_ledger_create_or_update_ledger_user_request(
            user_id=user_id,
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace_async
    async def get_user_defined_endpoint(self, **kwargs: Any) -> JSON:
        """Gets a user defined endpoint.

        Returns the user defined endpoint in the ACL instance.

        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "metadata": {
                        "endpoints": {
                            "str": {
                                "delete": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                },
                                "get": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                },
                                "patch": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                },
                                "put": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                }
                            }
                        }
                    },
                    "modules": {}
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_user_defined_endpoint_request(
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @overload
    async def create_user_defined_endpoint(
        self, bundle: JSON, *, content_type: str = "application/json", **kwargs: Any
    ) -> None:
        """Creates a user defined endpoint.

        Creates the user defined endpoint in the ACL instance.

        :param bundle: bundle parameter description. Required.
        :type bundle: JSON
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                bundle = {
                    "metadata": {
                        "endpoints": {
                            "str": {
                                "delete": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                },
                                "get": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                },
                                "patch": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                },
                                "put": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                }
                            }
                        }
                    },
                    "modules": {}
                }
        """

    @overload
    async def create_user_defined_endpoint(
        self, bundle: IO[bytes], *, content_type: str = "application/json", **kwargs: Any
    ) -> None:
        """Creates a user defined endpoint.

        Creates the user defined endpoint in the ACL instance.

        :param bundle: bundle parameter description. Required.
        :type bundle: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def create_user_defined_endpoint(self, bundle: Union[JSON, IO[bytes]], **kwargs: Any) -> None:
        """Creates a user defined endpoint.

        Creates the user defined endpoint in the ACL instance.

        :param bundle: bundle parameter description. Is either a JSON type or a IO[bytes] type.
         Required.
        :type bundle: JSON or IO[bytes]
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                bundle = {
                    "metadata": {
                        "endpoints": {
                            "str": {
                                "delete": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                },
                                "get": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                },
                                "patch": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                },
                                "put": {
                                    "authn_policies": [
                                        {}
                                    ],
                                    "forwarding_required": "str",
                                    "interpreter_reuse": {
                                        "key": "str"
                                    },
                                    "js_function": "str",
                                    "js_module": "str",
                                    "mode": "str",
                                    "openapi": {},
                                    "openapi_hidden": bool,
                                    "redirection_strategy": "str"
                                }
                            }
                        }
                    },
                    "modules": {}
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[None] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(bundle, (IOBase, bytes)):
            _content = bundle
        else:
            _json = bundle

        _request = build_confidential_ledger_create_user_defined_endpoint_request(
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [201]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @distributed_trace_async
    async def get_runtime_options(self, **kwargs: Any) -> JSON:
        """Runtime options for user defined endpoints.

        It returns the runtime options.

        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "log_exception_details": bool,
                    "max_cached_interpreters": 0,
                    "max_execution_time_ms": 0,
                    "max_heap_bytes": 0,
                    "max_stack_bytes": 0,
                    "return_exception_details": bool
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_runtime_options_request(
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @overload
    async def update_runtime_options(
        self, js_runtime_options: JSON, *, content_type: str = "application/json", **kwargs: Any
    ) -> JSON:
        """Runtime options for user defined endpoints.

        Updates the runtime options.

        :param js_runtime_options: JS runtime options. Required.
        :type js_runtime_options: JSON
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                js_runtime_options = {
                    "log_exception_details": bool,
                    "max_cached_interpreters": 0,
                    "max_execution_time_ms": 0,
                    "max_heap_bytes": 0,
                    "max_stack_bytes": 0,
                    "return_exception_details": bool
                }

                # response body for status code(s): 200
                response == {
                    "log_exception_details": bool,
                    "max_cached_interpreters": 0,
                    "max_execution_time_ms": 0,
                    "max_heap_bytes": 0,
                    "max_stack_bytes": 0,
                    "return_exception_details": bool
                }
        """

    @overload
    async def update_runtime_options(
        self, js_runtime_options: IO[bytes], *, content_type: str = "application/json", **kwargs: Any
    ) -> JSON:
        """Runtime options for user defined endpoints.

        Updates the runtime options.

        :param js_runtime_options: JS runtime options. Required.
        :type js_runtime_options: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "log_exception_details": bool,
                    "max_cached_interpreters": 0,
                    "max_execution_time_ms": 0,
                    "max_heap_bytes": 0,
                    "max_stack_bytes": 0,
                    "return_exception_details": bool
                }
        """

    @distributed_trace_async
    async def update_runtime_options(self, js_runtime_options: Union[JSON, IO[bytes]], **kwargs: Any) -> JSON:
        """Runtime options for user defined endpoints.

        Updates the runtime options.

        :param js_runtime_options: JS runtime options. Is either a JSON type or a IO[bytes] type.
         Required.
        :type js_runtime_options: JSON or IO[bytes]
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                js_runtime_options = {
                    "log_exception_details": bool,
                    "max_cached_interpreters": 0,
                    "max_execution_time_ms": 0,
                    "max_heap_bytes": 0,
                    "max_stack_bytes": 0,
                    "return_exception_details": bool
                }

                # response body for status code(s): 200
                response == {
                    "log_exception_details": bool,
                    "max_cached_interpreters": 0,
                    "max_execution_time_ms": 0,
                    "max_heap_bytes": 0,
                    "max_stack_bytes": 0,
                    "return_exception_details": bool
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[JSON] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(js_runtime_options, (IOBase, bytes)):
            _content = js_runtime_options
        else:
            _json = js_runtime_options

        _request = build_confidential_ledger_update_runtime_options_request(
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace_async
    async def get_user_defined_endpoints_module(self, *, module_name: str, **kwargs: Any) -> JSON:
        """Module for user defined endpoints.

        It gets the module for the user defined endpoint.

        :keyword module_name: module name of the user defined endpoint. Required.
        :paramtype module_name: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "module": "str",
                    "name": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_user_defined_endpoints_module_request(
            module_name=module_name,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace
    def list_user_defined_functions(self, **kwargs: Any) -> AsyncIterable[JSON]:
        """Retrieves a list of user defined functions present in the Confidential Ledger.

        User defined functions stored in the Confidential Ledger.

        :return: An iterator like instance of JSON object
        :rtype: ~azure.core.async_paging.AsyncItemPaged[JSON]
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "code": "str",
                    "id": "str"
                }
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                _request = build_confidential_ledger_list_user_defined_functions_request(
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                _request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                path_format_arguments = {
                    "ledgerEndpoint": self._serialize.url(
                        "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            return _request

        async def extract_data(pipeline_response):
            deserialized = pipeline_response.http_response.json()
            list_of_elem = deserialized.get("functions", [])
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.get("nextLink") or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            _request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
                _request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                raise HttpResponseError(response=response)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @distributed_trace_async
    async def delete_user_defined_function(self, function_id: str, **kwargs: Any) -> None:
        """Deletes a user defined function from the Confidential Ledger.

        Deletes a user defined function from the Confidential Ledger.

        :param function_id: Identifies a user defined function. Required.
        :type function_id: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[None] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_delete_user_defined_function_request(
            function_id=function_id,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [204]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @distributed_trace_async
    async def get_user_defined_function(self, function_id: str, **kwargs: Any) -> JSON:
        """Gets a user defined function.

        Returns the user defined function in the Confidential Ledger.

        :param function_id: Identifies a user defined function. Required.
        :type function_id: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "code": "str",
                    "id": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[JSON] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_user_defined_function_request(
            function_id=function_id,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @overload
    async def create_user_defined_function(
        self, function_id: str, user_defined_function: JSON, *, content_type: str = "application/json", **kwargs: Any
    ) -> JSON:
        """Creates a user defined function.

        Creates the user defined function in the Confidential Ledger.

        :param function_id: Identifies a user defined function. Required.
        :type function_id: str
        :param user_defined_function: Specify a user defined function of a Confidential Ledger.
         Required.
        :type user_defined_function: JSON
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                user_defined_function = {
                    "code": "str",
                    "id": "str"
                }

                # response body for status code(s): 200, 201
                response == {
                    "code": "str",
                    "id": "str"
                }
        """

    @overload
    async def create_user_defined_function(
        self,
        function_id: str,
        user_defined_function: IO[bytes],
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> JSON:
        """Creates a user defined function.

        Creates the user defined function in the Confidential Ledger.

        :param function_id: Identifies a user defined function. Required.
        :type function_id: str
        :param user_defined_function: Specify a user defined function of a Confidential Ledger.
         Required.
        :type user_defined_function: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200, 201
                response == {
                    "code": "str",
                    "id": "str"
                }
        """

    @distributed_trace_async
    async def create_user_defined_function(
        self, function_id: str, user_defined_function: Union[JSON, IO[bytes]], **kwargs: Any
    ) -> JSON:
        """Creates a user defined function.

        Creates the user defined function in the Confidential Ledger.

        :param function_id: Identifies a user defined function. Required.
        :type function_id: str
        :param user_defined_function: Specify a user defined function of a Confidential Ledger. Is
         either a JSON type or a IO[bytes] type. Required.
        :type user_defined_function: JSON or IO[bytes]
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                user_defined_function = {
                    "code": "str",
                    "id": "str"
                }

                # response body for status code(s): 200, 201
                response == {
                    "code": "str",
                    "id": "str"
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[JSON] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(user_defined_function, (IOBase, bytes)):
            _content = user_defined_function
        else:
            _json = user_defined_function

        _request = build_confidential_ledger_create_user_defined_function_request(
            function_id=function_id,
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200, 201]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        response_headers = {}
        response_headers["x-ms-ccf-transaction-id"] = self._deserialize(
            "str", response.headers.get("x-ms-ccf-transaction-id")
        )

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), response_headers)  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @overload
    async def execute_user_defined_function(
        self,
        function_id: str,
        user_defined_function_execution_properties: Optional[JSON] = None,
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> JSON:
        """Executes a user defined function.

        Executes the user defined function in the Confidential Ledger.

        :param function_id: Identifies a user defined function. Required.
        :type function_id: str
        :param user_defined_function_execution_properties: Specify user defined function execution
         properties. Default value is None.
        :type user_defined_function_execution_properties: JSON
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                user_defined_function_execution_properties = {
                    "arguments": [
                        "str"
                    ],
                    "exportedFunctionName": "str",
                    "runtimeOptions": {
                        "log_exception_details": bool,
                        "max_cached_interpreters": 0,
                        "max_execution_time_ms": 0,
                        "max_heap_bytes": 0,
                        "max_stack_bytes": 0,
                        "return_exception_details": bool
                    }
                }

                # response body for status code(s): 200
                response == {
                    "status": "str",
                    "error": {
                        "message": "str"
                    },
                    "result": {
                        "returnValue": "str"
                    }
                }
        """

    @overload
    async def execute_user_defined_function(
        self,
        function_id: str,
        user_defined_function_execution_properties: Optional[IO[bytes]] = None,
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> JSON:
        """Executes a user defined function.

        Executes the user defined function in the Confidential Ledger.

        :param function_id: Identifies a user defined function. Required.
        :type function_id: str
        :param user_defined_function_execution_properties: Specify user defined function execution
         properties. Default value is None.
        :type user_defined_function_execution_properties: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == {
                    "status": "str",
                    "error": {
                        "message": "str"
                    },
                    "result": {
                        "returnValue": "str"
                    }
                }
        """

    @distributed_trace_async
    async def execute_user_defined_function(
        self,
        function_id: str,
        user_defined_function_execution_properties: Optional[Union[JSON, IO[bytes]]] = None,
        **kwargs: Any
    ) -> JSON:
        """Executes a user defined function.

        Executes the user defined function in the Confidential Ledger.

        :param function_id: Identifies a user defined function. Required.
        :type function_id: str
        :param user_defined_function_execution_properties: Specify user defined function execution
         properties. Is either a JSON type or a IO[bytes] type. Default value is None.
        :type user_defined_function_execution_properties: JSON or IO[bytes]
        :return: JSON object
        :rtype: JSON
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                user_defined_function_execution_properties = {
                    "arguments": [
                        "str"
                    ],
                    "exportedFunctionName": "str",
                    "runtimeOptions": {
                        "log_exception_details": bool,
                        "max_cached_interpreters": 0,
                        "max_execution_time_ms": 0,
                        "max_heap_bytes": 0,
                        "max_stack_bytes": 0,
                        "return_exception_details": bool
                    }
                }

                # response body for status code(s): 200
                response == {
                    "status": "str",
                    "error": {
                        "message": "str"
                    },
                    "result": {
                        "returnValue": "str"
                    }
                }
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[JSON] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(user_defined_function_execution_properties, (IOBase, bytes)):
            _content = user_defined_function_execution_properties
        else:
            if user_defined_function_execution_properties is not None:
                _json = user_defined_function_execution_properties
            else:
                _json = None

        _request = build_confidential_ledger_execute_user_defined_function_request(
            function_id=function_id,
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(JSON, deserialized), {})  # type: ignore

        return cast(JSON, deserialized)  # type: ignore

    @distributed_trace_async
    async def get_user_defined_role(self, *, role_name: str, **kwargs: Any) -> List[JSON]:
        """Gets role actions for user defined roles.

        user defined roles allow users to define and manage app specific AuthZ policy.

        :keyword role_name: user defined role name. Required.
        :paramtype role_name: str
        :return: list of JSON object
        :rtype: list[JSON]
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # response body for status code(s): 200
                response == [
                    {
                        "roleActions": [
                            "str"
                        ],
                        "roleName": "str"
                    }
                ]
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[List[JSON]] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_get_user_defined_role_request(
            role_name=role_name,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if response.content:
            deserialized = response.json()
        else:
            deserialized = None

        if cls:
            return cls(pipeline_response, cast(List[JSON], deserialized), {})  # type: ignore

        return cast(List[JSON], deserialized)  # type: ignore

    @overload
    async def create_user_defined_role(
        self, roles: List[JSON], *, content_type: str = "application/json", **kwargs: Any
    ) -> None:
        """Creates new roles and their actions.

        User defined roles allow users to define and manage app specific AuthZ policy.

        :param roles: user defined role. Required.
        :type roles: list[JSON]
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                roles = [
                    {
                        "roleActions": [
                            "str"
                        ],
                        "roleName": "str"
                    }
                ]
        """

    @overload
    async def create_user_defined_role(
        self, roles: IO[bytes], *, content_type: str = "application/json", **kwargs: Any
    ) -> None:
        """Creates new roles and their actions.

        User defined roles allow users to define and manage app specific AuthZ policy.

        :param roles: user defined role. Required.
        :type roles: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def create_user_defined_role(self, roles: Union[List[JSON], IO[bytes]], **kwargs: Any) -> None:
        """Creates new roles and their actions.

        User defined roles allow users to define and manage app specific AuthZ policy.

        :param roles: user defined role. Is either a [JSON] type or a IO[bytes] type. Required.
        :type roles: list[JSON] or IO[bytes]
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[None] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(roles, (IOBase, bytes)):
            _content = roles
        else:
            _json = roles

        _request = build_confidential_ledger_create_user_defined_role_request(
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @overload
    async def update_user_defined_role(
        self, roles: List[JSON], *, content_type: str = "application/json", **kwargs: Any
    ) -> None:
        """Patch replaces the allowed action on existing roles,if the desire is to remove an existing
        action, the role must be deleted and recreated.

        User defined roles allow users to define and manage app specific AuthZ policy.

        :param roles: user defined role. Required.
        :type roles: list[JSON]
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:

        Example:
            .. code-block:: python

                # JSON input template you can fill out and use as your body input.
                roles = [
                    {
                        "roleActions": [
                            "str"
                        ],
                        "roleName": "str"
                    }
                ]
        """

    @overload
    async def update_user_defined_role(
        self, roles: IO[bytes], *, content_type: str = "application/json", **kwargs: Any
    ) -> None:
        """Patch replaces the allowed action on existing roles,if the desire is to remove an existing
        action, the role must be deleted and recreated.

        User defined roles allow users to define and manage app specific AuthZ policy.

        :param roles: user defined role. Required.
        :type roles: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def update_user_defined_role(self, roles: Union[List[JSON], IO[bytes]], **kwargs: Any) -> None:
        """Patch replaces the allowed action on existing roles,if the desire is to remove an existing
        action, the role must be deleted and recreated.

        User defined roles allow users to define and manage app specific AuthZ policy.

        :param roles: user defined role. Is either a [JSON] type or a IO[bytes] type. Required.
        :type roles: list[JSON] or IO[bytes]
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[None] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(roles, (IOBase, bytes)):
            _content = roles
        else:
            _json = roles

        _request = build_confidential_ledger_update_user_defined_role_request(
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @distributed_trace_async
    async def delete_user_defined_role(self, *, role_name: str, **kwargs: Any) -> None:
        """Deletes user defined roles.

        A user defined role allows the users to create and manage their own role actions using the API.

        :keyword role_name: user defined role name. Required.
        :paramtype role_name: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map: MutableMapping = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[None] = kwargs.pop("cls", None)

        _request = build_confidential_ledger_delete_user_defined_role_request(
            role_name=role_name,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        path_format_arguments = {
            "ledgerEndpoint": self._serialize.url(
                "self._config.ledger_endpoint", self._config.ledger_endpoint, "str", skip_quote=True
            ),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # type: ignore # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: use_collections_async.py
DESCRIPTION:
    This sample demonstrates how to use collections in Confidential Ledger. In this sample, we write
    ledger entries to different collections. Collections may be used to group semantically or
    logically related ledger entries.
USAGE:
    python use_collections_async.py
    Set the environment variables with your own values before running the sample:
    1) CONFIDENTIALLEDGER_ENDPOINT - the endpoint of the Confidential Ledger.
"""

import asyncio
import logging
import os
import sys
import tempfile
from typing import Any, Dict

from azure.confidentialledger.aio import ConfidentialLedgerClient
from azure.confidentialledger.certificate.aio import (
    ConfidentialLedgerCertificateClient,
)
from azure.identity.aio import DefaultAzureCredential


logging.basicConfig(level=logging.ERROR)
LOG = logging.getLogger()


async def main():
    # Set the values of the client ID, tenant ID, and client secret of the AAD application as
    # environment variables:
    #   AZURE_CLIENT_ID, AZURE_TENANT_ID, AZURE_CLIENT_SECRET, CONFIDENTIALLEDGER_ENDPOINT
    try:
        ledger_endpoint = os.environ["CONFIDENTIALLEDGER_ENDPOINT"]
    except KeyError:
        LOG.error(
            "Missing environment variable 'CONFIDENTIALLEDGER_ENDPOINT' - "
            "please set it before running the example"
        )
        sys.exit(1)

    # Under the current URI format, the ledger id is the first part of the ledger endpoint.
    # i.e. https://<ledger id>.confidential-ledger.azure.com
    ledger_id = ledger_endpoint.replace("https://", "").split(".")[0]

    identity_service_client = ConfidentialLedgerCertificateClient()  # type: ignore[call-arg]
    async with identity_service_client:
        ledger_certificate = await identity_service_client.get_ledger_identity(
            ledger_id
        )

    # The Confidential Ledger's TLS certificate must be written to a file to be used by the
    # ConfidentialLedgerClient. Here, we write it to a temporary file so that is is cleaned up
    # automatically when the program exits.
    with tempfile.TemporaryDirectory() as tempdir:
        ledger_cert_file = os.path.join(tempdir, f"{ledger_id}.pem")
        with open(ledger_cert_file, "w") as outfile:
            outfile.write(ledger_certificate["ledgerTlsCertificate"])

        print(
            f"Ledger certificate has been written to {ledger_cert_file}. "
            "It will be deleted when the script completes."
        )

        # Build a client through AAD
        credential = DefaultAzureCredential()
        ledger_client = ConfidentialLedgerClient(
            ledger_endpoint,
            credential=credential,
            ledger_certificate_path=ledger_cert_file,
        )

        # Using the async objects as a context manager ensures they are properly closed after use.
        async with credential:
            async with ledger_client:
                print("This Confidential Ledger will contain messages from different senders.")
                print(
                    "We will group ledger entries by the sender of the message. For all client "
                    "methods that take an optional 'collection_id' parameter, if none is provided, "
                    "a service-assigned, default collection id will be assigned."
                )

                tids: Dict[Any, Dict[str, Any]] = {}
                senders = [None, "Alice", "Bob"]
                for msg_idx in range(3):
                    for sender in senders:
                        if sender is None:
                            msg = f"My message {msg_idx}"
                        else:
                            msg = f"{sender}'s message {msg_idx}"

                        post_poller = await ledger_client.begin_create_ledger_entry(  # type: ignore[attr-defined]
                            entry={"contents": msg}, collection_id=sender,
                        )
                        post_result = await post_poller.result()

                        if sender is None:
                            print(
                                f"Wrote '{msg}' to the default collection at "
                                f"{post_result['transactionId']}"
                            )
                        else:
                            print(
                                f"Wrote '{msg}' to collection {sender} at "
                                f"{post_result['transactionId']}"
                            )

                        if sender not in tids:
                            tids[sender] = {}
                            tids[sender]["first"] = post_result["transactionId"]

                        tids[sender]["last"] = post_result["transactionId"]

                print("Let's retrieve the latest entry in each collection")
                for sender in senders:
                    current_entry = await ledger_client.get_current_ledger_entry()
                    
                    output = "Current entry in {0} is {1}"
                    print(
                        output.format(
                            "default collection" if sender is None else f"{sender}'s collection",
                            current_entry["contents"],
                        )
                    )

                print("Let's retrieve the first entry in each collection")
                for sender in senders:
                    get_poller = await ledger_client.begin_get_ledger_entry(  # type: ignore[attr-defined]
                        tids[sender]["first"],
                        collection_id=sender
                    )
                    first_entry = await get_poller.result()
                    
                    output = "First entry in {0} is {1}"
                    print(
                        output.format(
                            "default collection" if sender is None else f"{sender}'s collection",
                            first_entry["entry"]["contents"],
                        )
                    )

                print("Let's retrieve get all the entries in each collection")
                for sender in senders:
                    entries_list = ledger_client.list_ledger_entries(
                        collection_id=sender,
                        from_transaction_id=tids[sender]["first"],
                        to_transaction_id=tids[sender]["last"],
                    )
                    
                    async for entry in entries_list:
                        output = "Entry in {0}: {1}"
                        print(
                            output.format(
                                "default collection" if sender is None else f"{sender}'s "
                                "collection",
                                entry,
                            )
                        )

                collection_ids = []
                collections = ledger_client.list_collections()
                async for collection in collections:
                    collection_ids.append(collection["collectionId"])

                print(
                    "In conclusion, these are all the collections in the Confidential Ledger:\n" +
                    "\n\t".join(collection_ids)
                )


if __name__ == "__main__":
    asyncio.run(main())



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
"""Models for application claims."""

from typing import Any, Dict, Optional, Union
from dataclasses import dataclass


@dataclass
class LedgerEntryClaim:
    """
    LedgerEntryClaim represents an Application Claim derived from ledger entry data.

    :keyword protocol: The protocol used to compute the claim.
    :paramtype protocol: str

    :keyword collectionId: The collection ID of the ledger entry.
    :paramtype collectionId: str

    :keyword contents: The contents of the ledger entry.
    :paramtype contents: str

    :keyword secretKey: The secret key used to compute the claim digest.
    :paramtype secretKey: str
    """

    protocol: str
    collectionId: str
    contents: str
    secretKey: str

    @classmethod
    def from_dict(cls, ledger_entry_claim_dict: Dict[str, Any]):
        """Create a new instance of this class from a dictionary.

        :param dict[str, any] ledger_entry_claim_dict: The dictionary representation of the ledger entry claim.
        :return: A new instance of this class corresponding to the provided dictionary.
        :rtype: LedgerEntryClaim
        """

        return cls(**ledger_entry_claim_dict)


@dataclass
class ClaimDigest:
    """
    ClaimDigest represents an Application Claim in digested form.

    :keyword protocol: The protocol used to compute the claim.
    :paramtype protocol: str

    :keyword value: The digest of the claim.
    :paramtype value: str
    """

    protocol: str
    value: str

    @classmethod
    def from_dict(cls, ledger_entry_claim_dict: Dict[str, Any]):
        """Create a new instance of this class from a dictionary.

        :param dict[str, any] ledger_entry_claim_dict: The dictionary representation of the claim digest.
        :return: A new instance of this class corresponding to the provided dictionary.
        :rtype: ClaimDigest
        """

        return cls(**ledger_entry_claim_dict)


@dataclass
class ApplicationClaim:
    """
    ApplicationClaim represents a claim of a ledger application.

    :keyword kind: The kind of the claim.
    :paramtype kind: str

    :keyword ledgerEntry: The ledger entry claim.
    :paramtype ledgerEntry: Optional[Union[Dict[str, Any], LedgerEntryClaim]]

    :keyword digest: The claim digest object.
    :paramtype digest: Optional[Union[Dict[str, Any], ClaimDigest]]
    """

    kind: str
    ledgerEntry: Optional[LedgerEntryClaim] = None
    digest: Optional[ClaimDigest] = None

    def __init__(
        self,
        kind: str,
        ledgerEntry: Optional[Union[Dict[str, Any], LedgerEntryClaim]] = None,
        digest: Optional[Union[Dict[str, Any], ClaimDigest]] = None,
        **kwargs: Any
    ):
        """
        :keyword kind: The kind of the claim.
        :paramtype kind: str

        :keyword ledgerEntry: The ledger entry claim.
        :paramtype ledgerEntry: Optional[Union[Dict[str, Any], LedgerEntryClaim]]

        :keyword digest: The claim digest object.
        :paramtype digest: Optional[Union[Dict[str, Any], ClaimDigest]]
        """
        self.kind = kind

        if ledgerEntry:
            if isinstance(ledgerEntry, LedgerEntryClaim):
                self.ledgerEntry = ledgerEntry
            else:
                self.ledgerEntry = LedgerEntryClaim.from_dict(ledgerEntry)
        else:
            self.ledgerEntry = None

        if digest:
            if isinstance(digest, ClaimDigest):
                self.digest = digest
            else:
                self.digest = ClaimDigest.from_dict(digest)
        else:
            self.digest = None

        self.kwargs = kwargs

    @classmethod
    def from_dict(cls, claim_dict: Dict[str, Any]):
        """Create a new instance of this class from a dictionary.

        :param dict[str, any] claim_dict: The dictionary representation of the application claim.
        :return: A new instance of this class corresponding to the provided dictionary.
        :rtype: ApplicationClaim
        """

        return cls(**claim_dict)



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
"""Algorithm implementation for verifying Azure Confidential Ledger write
transaction receipts."""

from base64 import b64decode
from hashlib import sha256
from typing import Dict, List, Any, cast, Optional

from cryptography.x509 import load_pem_x509_certificate, Certificate
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, utils
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from azure.confidentialledger.receipt._receipt_models import (
    LeafComponents,
    ProofElement,
    Receipt,
)

from azure.confidentialledger.receipt._utils import (
    _convert_dict_to_camel_case,
)

from azure.confidentialledger.receipt._claims_digest_computation import (
    compute_claims_digest,
)


def verify_receipt(
    receipt: Dict[str, Any],
    service_cert: str,
    *,
    application_claims: Optional[List[Dict[str, Any]]] = None,
) -> None:
    """Verify that a given Azure Confidential Ledger write transaction receipt
    is valid from its content and the Confidential Ledger service identity
    certificate.

    :param receipt: Receipt dictionary containing the content of an Azure
     Confidential Ledger write transaction receipt.
    :type receipt: Dict[str, Any]

    :param service_cert: String containing the PEM-encoded
     certificate of the Confidential Ledger service identity.
    :type service_cert: str

    :keyword application_claims: List of application claims to be verified against the receipt.
    :paramtype application_claims: Optional[List[Dict[str, Any]]]

    :raises ValueError: If the receipt verification has failed.
    """

    # Validate receipt content and convert it into a Receipt model
    receipt_obj = _preprocess_input_receipt(receipt)

    # Validate application claims provided by the user, if any
    if application_claims:
        computed_claims_digest = compute_claims_digest(application_claims)
        if computed_claims_digest != receipt_obj.leafComponents.claimsDigest:
            raise ValueError(
                "The computed claims digest from application claims does not match the receipt claims digest."
            )

    # Load node PEM certificate
    node_cert = _load_and_verify_pem_certificate(receipt_obj.cert)

    # Verify node certificate is endorsed by the service certificate
    # through endorsements certificates
    _verify_node_cert_endorsed_by_service_cert(
        node_cert, service_cert, cast(List[str], receipt_obj.serviceEndorsements)
    )

    # Compute hash of the leaf node in the Merkle Tree corresponding
    # to the transaction associated to the given receipt
    leaf_node_hash = _compute_leaf_node_hash(receipt_obj.leafComponents)

    # Compute root of the Merkle Tree at the time the transaction was committed
    root_node_hash = _compute_root_node_hash(leaf_node_hash, receipt_obj.proof)

    # Verify signature of the signing node over the root of the tree with
    # node certificate public key
    _verify_signature_over_root_node_hash(
        receipt_obj.signature, node_cert, cast(str, receipt_obj.nodeId), root_node_hash
    )


def _preprocess_input_receipt(receipt_dict: Dict[str, Any]) -> Receipt:
    """Preprocess input receipt dictionary, validate its content, and returns a
    valid Receipt object based on the vetted input data.

    :param dict[str, any] receipt_dict: Receipt dictionary
    :return: Receipt object
    :rtype: Receipt
    """

    # Convert any key in the receipt dictionary to camel case
    # to match the model fields (we do this because customers may
    # provide receipts with snake case keys since they were returned
    # by older ACL instances)
    receipt_dict = _convert_dict_to_camel_case(receipt_dict)

    _validate_receipt_content(receipt_dict)

    # Convert receipt JSON object to Receipt model
    return Receipt.from_dict(receipt_dict)


def _validate_receipt_content(receipt: Dict[str, Any]):
    """Validate the content of a write transaction receipt.

    :param dict[str, any] receipt: Receipt dictionary
    """

    try:
        assert "cert" in receipt
        assert isinstance(receipt["cert"], str)

        assert "leafComponents" in receipt
        assert isinstance(receipt["leafComponents"], dict)

        assert "claimsDigest" in receipt["leafComponents"]
        assert isinstance(receipt["leafComponents"]["claimsDigest"], str)

        assert "commitEvidence" in receipt["leafComponents"]
        assert isinstance(receipt["leafComponents"]["commitEvidence"], str)

        assert "writeSetDigest" in receipt["leafComponents"]
        assert isinstance(receipt["leafComponents"]["writeSetDigest"], str)

        assert "proof" in receipt
        assert isinstance(receipt["proof"], list)

        # Validate elements in proof
        for elem in receipt["proof"]:
            assert "left" in elem or "right" in elem
            if "left" in elem:
                assert isinstance(elem["left"], str)
            if "right" in elem:
                assert isinstance(elem["right"], str)

        assert "signature" in receipt
        assert isinstance(receipt["signature"], str)

        # Validate nodeId, if present
        if "nodeId" in receipt:
            assert isinstance(receipt["nodeId"], str)

        # Validate serviceEndorsements, if present
        if "serviceEndorsements" in receipt:
            assert isinstance(receipt["serviceEndorsements"], list)

            # Validate elements in serviceEndorsements
            for elem in receipt["serviceEndorsements"]:
                assert isinstance(elem, str)

    except Exception as exception:
        raise ValueError("The receipt content is invalid.") from exception


def _verify_signature_over_root_node_hash(
    signature: str, node_cert: Certificate, node_id: str, root_node_hash: bytes
) -> None:
    """Verify signature over root node hash of the Merkle Tree using node
    certificate public key.

    :param str signature: Signature
    :param Certificate node_cert: Node certificate
    :param str node_id: Node ID
    :param bytes root_node_hash: Root node hash
    """

    try:
        # Verify public key contained in the node certificate is equal to the node_id
        public_key_bytes = node_cert.public_key().public_bytes(
            Encoding.DER, PublicFormat.SubjectPublicKeyInfo
        )

        if node_id is not None:
            assert sha256(public_key_bytes).digest() == bytes.fromhex(node_id)

        # Verify signature over root node hash using node certificate public key
        _verify_ec_signature(
            node_cert,
            b64decode(signature, validate=True),
            root_node_hash,
            hashes.SHA256(),
        )

    except Exception as exception:
        raise ValueError(
            f"Encountered exception when verifying signature {signature} over root node hash."
        ) from exception


def _compute_leaf_node_hash(leaf_components: LeafComponents) -> bytes:
    """Compute the hash of the leaf node associated to a transaction given the
    leaf components from a write transaction receipt.

    :param LeafComponents leaf_components: Leaf components
    :return: Leaf node hash
    :rtype: bytes
    """

    try:
        # Digest commit evidence string
        commit_evidence_digest = sha256(
            leaf_components.commitEvidence.encode()
        ).digest()

        # Convert write set digest to bytes
        write_set_digest = bytes.fromhex(leaf_components.writeSetDigest)

        # Convert claims digest to bytes
        claims_digest = bytes.fromhex(leaf_components.claimsDigest)

        # Create leaf node hash by hashing the concatenation of its three components
        # as bytes objects in the following order:
        # 1. write_set_digest
        # 2. commit_evidence_digest
        # 3. claims_digest
        return sha256(
            write_set_digest + commit_evidence_digest + claims_digest
        ).digest()

    except Exception as exception:
        raise ValueError(
            f"Encountered exception when computing leaf node hash from leaf components {leaf_components}."
        ) from exception


def _compute_root_node_hash(leaf_hash: bytes, proof: List[ProofElement]) -> bytes:
    """Re-compute the hash of the root of the Merkle tree from a leaf node hash
    and a receipt proof list containing the required nodes hashes for the
    computation.

    :param bytes leaf_hash: Leaf node hash
    :param list[ProofElement] proof: Receipt proof list
    :return: Root node hash
    :rtype: bytes
    """

    try:
        # Initialize current hash to leaf hash
        current_node_hash = leaf_hash

        # Iterate through all the elements in proof list
        for element in proof:
            # Check that the current element only contains either one left or right node hash
            if (
                element is None
                or (element.left is None and element.right is None)
                or (element.left is not None and element.right is not None)
            ):
                raise ValueError(
                    "Invalid proof element in receipt: element must contain either one left or right node hash."
                )

            parent_node_hash = b""

            # If the current element contains a left hash, concatenate the left hash and the current node hash
            if element.left is not None:
                parent_node_hash = bytes.fromhex(element.left) + current_node_hash

            # If the current element contains a right hash, concatenate the current node hash and the right hash
            if element.right is not None:
                parent_node_hash = current_node_hash + bytes.fromhex(element.right)

            # Hash the parent node hash
            current_node_hash = sha256(parent_node_hash).digest()

        return current_node_hash

    except Exception as exception:
        raise ValueError(
            f"Encountered exception when computing root node hash from proof list {proof}."
        ) from exception


def _verify_certificate_endorsement(
    endorsee: Certificate, endorser: Certificate
) -> None:
    """Verify that the endorser certificate has endorsed endorsee
    certificate using ECDSA.

    :param Certificate endorsee: Endorsee certificate
    :param Certificate endorser: Endorser certificate
    """

    try:
        # Extract TBS certificate hash from endorsee certificate
        hash_algorithm = cast(hashes.HashAlgorithm, endorsee.signature_hash_algorithm)
        digester = hashes.Hash(hash_algorithm)
        digester.update(endorsee.tbs_certificate_bytes)
        cert_digest = digester.finalize()

        # Verify endorser signature over endorsee certificate digest
        _verify_ec_signature(endorser, endorsee.signature, cert_digest, hash_algorithm)

    except Exception as exception:
        raise ValueError(
            f"Encountered exception when verifying endorsement of certificate {endorsee} by certificate {endorser}."
        ) from exception


def _verify_ec_signature(
    certificate: Certificate,
    signature: bytes,
    data: bytes,
    hash_algorithm: hashes.HashAlgorithm,
) -> None:
    """Verify a signature over data using the certificate public key.

    :param Certificate certificate: Certificate
    :param bytes signature: Signature
    :param bytes data: Data
    :param hashes.HashAlgorithm hash_algorithm: Hash algorithm
    """

    public_key = cast(ec.EllipticCurvePublicKey, certificate.public_key())

    public_key.verify(
        signature,
        data,
        ec.ECDSA(utils.Prehashed(hash_algorithm)),
    )


def _verify_node_cert_endorsed_by_service_cert(
    node_cert: Certificate, service_cert_str: str, endorsements_certs: List[str]
) -> None:
    """Check a node certificate is endorsed by a service certificate.

    If a list of endorsements certificates is not empty, check that the
    node certificate is transitively endorsed by the service certificate
    through the endorsement certificates in the list.

    :param Certificate node_cert: Node certificate
    :param str service_cert_str: Service certificate string
    :param list[str] endorsements_certs: Endorsements certificates list
    """

    current_cert = node_cert

    # Validate endorsement certificates list is present
    if endorsements_certs is None:
        endorsements_certs = []

    # Add service certificate to the list of endorsements certificates
    endorsements_certs.append(service_cert_str)

    # Iterate through all the endorsements certificates
    for endorsement in endorsements_certs:
        # Load endorsement PEM certificate
        endorsement_cert = _load_and_verify_pem_certificate(endorsement)

        # Verify endorsement certificate has endorsed current certificate
        _verify_certificate_endorsement(current_cert, endorsement_cert)

        # Set current certificate to endorsement certificate to continue the chain verification
        current_cert = endorsement_cert


def _load_and_verify_pem_certificate(cert_str: str) -> Certificate:
    """Load PEM certificate from a string representation and verify it is a
    valid certificate with expected Elliptic Curve public key.

    :param str cert_str: PEM certificate string
    :return: Certificate
    :rtype: Certificate
    """

    try:
        # Load certificate from string
        cert = load_pem_x509_certificate(cert_str.encode())

        # Verify public key is of the correct type
        assert isinstance(cert.public_key(), ec.EllipticCurvePublicKey)

        return cert

    except Exception as exception:
        raise ValueError(f"PEM certificate {cert_str} is not valid.") from exception
